/* drivers/video/sunxi/disp2/disp/lcd/jd9852.c
 *
 * Copyright (c) 2021 IAMLIUBO <https://github.com/imliubo>
 * 
 * Origin:
 * Copyright (c) 2017 Allwinnertech Co., Ltd.
 * Author: zhengxiaobin <zhengxiaobin@allwinnertech.com>
 *
 * jd9852 panel driver based on jd9852 driver
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2 as
 * published by the Free Software Foundation.
 &lcd0 {
	lcd_used            = <1>;

	lcd_driver_name     = "jd9852";
	lcd_backlight       = <100>;
	lcd_if              = <4>;

	lcd_x               = <240>;
	lcd_y               = <320>;
	lcd_width           = <30>;
	lcd_height          = <40>;
	lcd_dclk_freq       = <7>;

	lcd_pwm_used        = <1>;
	lcd_pwm_ch          = <2>;
	lcd_pwm_freq        = <1000>;
	lcd_pwm_pol         = <0>;
	lcd_pwm_max_limit   = <255>;

	lcd_hbp             = <20>;
	lcd_ht              = <304>;
	lcd_hspw            = <4>;
	lcd_vbp             = <6>;
	lcd_vt              = <336>;
	lcd_vspw            = <2>;

	lcd_dsi_if          = <0>;
	lcd_dsi_lane        = <1>;
	lcd_lvds_if         = <0>;
	lcd_lvds_colordepth = <0>;
	lcd_lvds_mode       = <0>;
	lcd_frm             = <0>;
	lcd_hv_clk_phase    = <0>;
	lcd_hv_sync_polarity= <0>;
	lcd_io_phase        = <0x0000>;
	lcd_gamma_en        = <0>;
	lcd_bright_curve_en = <0>;
	lcd_cmap_en         = <0>;
	lcd_fsync_en        = <0>;
	lcd_fsync_act_time  = <1000>;
	lcd_fsync_dis_time  = <1000>;
	lcd_fsync_pol       = <0>;

	deu_mode            = <0>;
	lcdgamma4iep        = <22>;
	smart_color         = <90>;

	lcd_gpio_0 =  <&pio PG 13 GPIO_ACTIVE_HIGH>;
	pinctrl-0 = <&dsi4lane_pins_a>;
	pinctrl-1 = <&dsi4lane_pins_b>;
};
*/
#include "jd9852.h"

static void lcd_power_on(u32 sel);
static void lcd_power_off(u32 sel);
static void lcd_bl_open(u32 sel);
static void lcd_bl_close(u32 sel);

static void lcd_panel_init(u32 sel);
static void lcd_panel_exit(u32 sel);

#define panel_reset(sel, val) sunxi_lcd_gpio_set_value(sel, 0, val)

static void lcd_cfg_panel_info(struct panel_extend_para *info)
{
	u32 i = 0, j = 0;
	u32 items;
	u8 lcd_gamma_tbl[][2] = {
		{0, 0},
		{15, 15},
		{30, 30},
		{45, 45},
		{60, 60},
		{75, 75},
		{90, 90},
		{105, 105},
		{120, 120},
		{135, 135},
		{150, 150},
		{165, 165},
		{180, 180},
		{195, 195},
		{210, 210},
		{225, 225},
		{240, 240},
		{255, 255},
	};

	u32 lcd_cmap_tbl[2][3][4] = {
		{
			{LCD_CMAP_G0, LCD_CMAP_B1, LCD_CMAP_G2, LCD_CMAP_B3},
			{LCD_CMAP_B0, LCD_CMAP_R1, LCD_CMAP_B2, LCD_CMAP_R3},
			{LCD_CMAP_R0, LCD_CMAP_G1, LCD_CMAP_R2, LCD_CMAP_G3},
		},
		{
			{LCD_CMAP_B3, LCD_CMAP_G2, LCD_CMAP_B1, LCD_CMAP_G0},
			{LCD_CMAP_R3, LCD_CMAP_B2, LCD_CMAP_R1, LCD_CMAP_B0},
			{LCD_CMAP_G3, LCD_CMAP_R2, LCD_CMAP_G1, LCD_CMAP_R0},
		},
	};

	items = sizeof(lcd_gamma_tbl) / 2;
	for (i = 0; i < items - 1; i++) {
		u32 num = lcd_gamma_tbl[i + 1][0] - lcd_gamma_tbl[i][0];

		for (j = 0; j < num; j++) {
			u32 value = 0;

			value = lcd_gamma_tbl[i][1] +
				((lcd_gamma_tbl[i + 1][1] - lcd_gamma_tbl[i][1])
				 * j) / num;
			info->lcd_gamma_tbl[lcd_gamma_tbl[i][0] + j] =
				(value << 16)
				+ (value << 8) + value;
		}
	}
	info->lcd_gamma_tbl[255] = (lcd_gamma_tbl[items - 1][1] << 16) +
				   (lcd_gamma_tbl[items - 1][1] << 8)
				   + lcd_gamma_tbl[items - 1][1];

	memcpy(info->lcd_cmap_tbl, lcd_cmap_tbl, sizeof(lcd_cmap_tbl));

}

static s32 lcd_open_flow(u32 sel)
{
	LCD_OPEN_FUNC(sel, lcd_power_on, 10);
	LCD_OPEN_FUNC(sel, lcd_panel_init, 10);
	LCD_OPEN_FUNC(sel, sunxi_lcd_tcon_enable, 50);
	LCD_OPEN_FUNC(sel, lcd_bl_open, 0);

	return 0;
}

static s32 lcd_close_flow(u32 sel)
{
	LCD_CLOSE_FUNC(sel, lcd_bl_close, 0);
	LCD_CLOSE_FUNC(sel, sunxi_lcd_tcon_disable, 0);
	LCD_CLOSE_FUNC(sel, lcd_panel_exit, 200);
	LCD_CLOSE_FUNC(sel, lcd_power_off, 500);

	return 0;
}

static void lcd_power_on(u32 sel)
{
	sunxi_lcd_pin_cfg(sel, 1);
	sunxi_lcd_delay_ms(50);
	panel_reset(sel, 1);
	sunxi_lcd_delay_ms(5);
	panel_reset(sel, 0);
	sunxi_lcd_delay_ms(10);
	panel_reset(sel, 1);
	sunxi_lcd_delay_ms(120);

}

static void lcd_power_off(u32 sel)
{
	sunxi_lcd_pin_cfg(sel, 0);
	sunxi_lcd_delay_ms(20);
	panel_reset(sel, 0);
	sunxi_lcd_delay_ms(5);
}

static void lcd_bl_open(u32 sel)
{
	sunxi_lcd_pwm_enable(sel);
}

static void lcd_bl_close(u32 sel)
{
	sunxi_lcd_pwm_disable(sel);
}

#define REGFLAG_DELAY 0XFC
#define REGFLAG_END_OF_TABLE 0xFD /* END OF REGISTERS MARKER */

struct LCM_setting_table {
	u8 cmd;
	u32 count;
	u8 para_list[32];
};

static struct LCM_setting_table lcm_jd9852_setting[] = {
    // PASSWORD
	{0xDF, 3, {0x98, 0x51, 0xE9}},
	{0xDE, 1, {0x00}},
    // VGMP,VGSP,VGMN,VGSN 4.2
	{0xB7, 4, {0x1E, 0x7D, 0x1E, 0x2B}},
    // Set_R_GAMMA
	{0xC8, 32, {0x3F, 0x31, 0x29, 0x28, 0x2C, 0x31, 0x2D, 0x2D, 0x2A, 0x28, 0x23, 0x16, 0x10, 0x0E, 0x08, 0x0E, 0x3F, 0x31, 0x29, 0x28, 0x2C, 0x31, 0x2D, 0x2C, 0x2A, 0x28, 0x23, 0x16, 0x10, 0x0E, 0x08, 0x0E}},
	// POW_CTRL
    {0xB9, 3, {0x33, 0x08, 0xCC}},
    // DCDC_SEL 
	{0xBB, 8, {0x46, 0x7A, 0x30, 0xD0, 0x7C, 0x60, 0x70, 0x70}},
    // VDDD_CTRL 
	{0xBC, 2, {0x38, 0x3C}},
    // SETSTBA
	{0xC0, 2, {0x31, 0x20}},
    // SETPANEL(default)  
	{0xC1, 1, {0x12}},
    // SETRGBCYC 
	{0xC3, 9, {0x08, 0x00, 0x0A, 0x10, 0x08, 0x54, 0x45, 0x71, 0x2C}},
    // SETRGBCYC(default)  
	{0xC4, 17, {0x00, 0xA0, 0x79, 0x0E, 0x0A, 0x16, 0x79, 0x0E, 0x0A, 0x16, 0x79, 0x0E, 0x0A, 0x16, 0x82, 0x00, 0x03}},
    // SET_GD(default)
	{0xD0, 6, {0x04, 0x0C, 0x6B, 0x0F, 0x07, 0x03}},
    // RAMCTRL(default)
	{0xD7, 2, {0x13, 0x00}},
	{0xDE, 1, {0x02}},
	{REGFLAG_DELAY, REGFLAG_DELAY, {1} },

    // DCDC_SET
	{0xB8, 5, {0x1D, 0xA0, 0x2F, 0x04, 0x33}},
    // SETRGBCYC2
	{0xC1, 4, {0x10, 0x66, 0x66, 0x01}},
    // OSCM
	{0xDE, 1, {0x00}},
	{0x11, 0, {}},
	{REGFLAG_DELAY, REGFLAG_DELAY, {120} },

	{0xDE, 1, {0x02}},
	{REGFLAG_DELAY, REGFLAG_DELAY, {1} },

	{0xC5, 3, {0x4E, 0x00, 0x00}},
	{REGFLAG_DELAY, REGFLAG_DELAY, {1} },

    // SETMIPI_2
	{0xCA, 3, {0x30, 0x20, 0xF4}},
	{REGFLAG_DELAY, REGFLAG_DELAY, {1} },

	{0xDE, 1, {0x04}},
	{REGFLAG_DELAY, REGFLAG_DELAY, {1} },

    // SETPHY3
	{0xD3, 1, {0x3C}},
	{REGFLAG_DELAY, REGFLAG_DELAY, {1} },


	{0xDE, 1, {0x00}},
	{REGFLAG_DELAY, REGFLAG_DELAY, {1} },

    // SLPOUT
	{0x29, 0, {} },
	{REGFLAG_DELAY, REGFLAG_DELAY, {120} },

	{REGFLAG_END_OF_TABLE, REGFLAG_END_OF_TABLE, {} }
};

static void lcd_panel_init(u32 sel)
{
	__u32 i;
	sunxi_lcd_dsi_clk_enable(sel);
	sunxi_lcd_delay_ms(100);
	for (i = 0;; i++) {
		if (lcm_jd9852_setting[i].count == REGFLAG_END_OF_TABLE)
			break;
		else if (lcm_jd9852_setting[i].count == REGFLAG_DELAY) {
			sunxi_lcd_delay_ms(lcm_jd9852_setting[i].para_list[0]);
		} else {
			dsi_gen_wr(sel, lcm_jd9852_setting[i].cmd,
				   lcm_jd9852_setting[i].para_list,
				   lcm_jd9852_setting[i].count);
		}
		/* break; */
	}
}

static void lcd_panel_exit(u32 sel)
{
	sunxi_lcd_dsi_dcs_write_0para(sel, 0x10);
	sunxi_lcd_delay_ms(80);
	sunxi_lcd_dsi_dcs_write_0para(sel, 0x28);
	sunxi_lcd_delay_ms(50);
}

/*sel: 0:lcd0; 1:lcd1*/
static s32 lcd_user_defined_func(u32 sel, u32 para1, u32 para2, u32 para3)
{
	return 0;
}

struct __lcd_panel jd9852_panel = {
	/* panel driver name, must mach the name of
	 * lcd_drv_name in sys_config.fex
	 */
	.name = "jd9852",
	.func = {
		.cfg_panel_info = lcd_cfg_panel_info,
		.cfg_open_flow = lcd_open_flow,
		.cfg_close_flow = lcd_close_flow,
		.lcd_user_defined_func = lcd_user_defined_func,
	},
};
